!=============================================================================!
!=============================================================================!
!!*                                                                         *!!
!!                 AstroGK Index for Heating Diagnostics Module              !!
!!                                   2008                                    !!
!!*                                                                         *!!
!=============================================================================!
!=============================================================================!
! Here are defined the indices for each calculated heating and energy quantity


module agk_heating_index
  implicit none
  private

! Quantities independent of nspec-------------------------------------------

! Energy quantities
  integer,parameter :: lenergy=1          !Total Energy (W)
  integer,parameter :: leapar=2           !Total Bperp energy (A_parallel)
  integer,parameter :: lebpar=3           !Total Bpar energy

! Heating Quantities (rate of change of energy)
  integer,parameter :: lantenna=4         !Total Antenna (input) energy
  integer,parameter :: lenergy_dot=5      !Total rate of change of energy (dW/dt)

! Balance Quantities
  integer,parameter :: lpowbal=6          !Power Balance
  integer,parameter :: lpowbalcoll=7      !Collisional Power Balance

! Quantities dependent on nspec---------------------------------------------

! Balance Quantities  
  integer, dimension(:), allocatable :: lentbal       !Entropy Balance

! Energy quantities
  integer, dimension(:), allocatable :: ldelfs2       !int T/F0 dfs^2/2
  integer, dimension(:), allocatable :: lhs2          !int T/F0  hs^2/2
  integer, dimension(:), allocatable :: lphis2        !int  q^2 n/T phi^2/2

! Heating Quantities (rate of change of energy)
  integer, dimension(:), allocatable :: limp_colls    !Implicit heating
  integer, dimension(:), allocatable :: lhypercoll    !Hypercollisional heating
  integer, dimension(:), allocatable :: lcollisions   !Collisional heating
  integer, dimension(:), allocatable :: lheating      !Dist Func heating input
  integer, dimension(:), allocatable :: lgradients    !Gradient heating

! Other useful numbers
  integer :: lemax                         !Numbers of heating quantities
  logical :: initialized = .false.

  public :: lenergy,leapar,lebpar,ldelfs2,lhs2,lphis2
  public :: lpowbal,lpowbalcoll,lentbal
  public :: lantenna,lenergy_dot
  public :: limp_colls,lhypercoll,lcollisions,lheating,lgradients
  public :: lemax

  public :: init_heating_index,finish_heating_index

contains
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------

! Initialize indices for all nspec dependent quantities

  subroutine init_heating_index(nspec)
    implicit none
! Passed
    integer :: nspec                   !Number of species
! Local
    integer :: n,i

    if (initialized) return
    initialized = .true.
    
! Allocate nspec size for all dependent quantities

    allocate(lentbal(1:nspec))
    allocate(ldelfs2(1:nspec))
    allocate(lhs2(1:nspec))
    allocate(lphis2(1:nspec))
    allocate(limp_colls(1:nspec))
    allocate(lhypercoll(1:nspec))
    allocate(lcollisions(1:nspec))
    allocate(lheating(1:nspec))
    allocate(lgradients(1:nspec))

! Set indices for these quantities

    n=lpowbalcoll
    do i=1,nspec
       lentbal(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       ldelfs2(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lhs2(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lphis2(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       limp_colls(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lhypercoll(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lcollisions(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lheating(i)=n+i
    enddo

    n=n+nspec
    do i=1,nspec
       lgradients(i)=n+i
    enddo

    n=n+nspec
    lemax=n

  end subroutine init_heating_index
!------------------------------------------------------------------------------
!                              AstroGK, 2008
!------------------------------------------------------------------------------
! Deallocate heating indices

  subroutine finish_heating_index
    implicit none

    !NOTE: There seems to be some problem with the code crashing here GGH 02APR08

! Deallocate all dependent quantities
    deallocate(lentbal,ldelfs2,lhs2,lphis2,limp_colls)
    deallocate(lhypercoll,lcollisions,lheating,lgradients)

  end subroutine finish_heating_index
!------------------------------------------------------------------------------
end module agk_heating_index
