# include "define.inc"

module file_utils

  implicit none
  private

  public :: init_file_utils
  ! subroutine init_file_utils (list, input, error, name)
  ! logical, intent (out) :: list
  ! logical, intent (in), optional :: input, error
  ! character(*), intent (in), optional :: name
  !   default: INPUT=.true., ERROR=.true., NAME="unknown"
  !   Set up run_name(s) and list_name for output files
  !   Open input file and strip comments, unless disabled with INPUT=.false.
  !   Open error output file, unless disabled with ERROR=.false.

  public :: init_job_name
  ! subroutine ...

  public :: finish_file_utils
  ! subroutine finish_file_utils
  !   Clean up files opened in init

  public :: trin
  ! logical :: trin
  ! flag denoting if transport input file present

  public :: run_name
  ! character(500) :: run_name
  !    Label for the run, taken from the command line

  public :: list_name
  ! character(500) :: list_name
  !    Label for the list, taken from the command line

!  public :: trin_name
  ! character(500) :: trin_name
  !    Label for the transport input file, taken from the command line

  public :: input_unit
  ! function input_unit (nml)
  ! character(*), intent (in) :: nml
  ! integer :: input_unit
  !    Rewind the input file to start of namelist NML,
  !    and return its unit number

  public :: input_unit_exist
  ! function input_unit_exist (nml,exist)
  ! character(*), intent (in) :: nml
  ! integer :: input_unit
  !    Rewind the input file to start of namelist NML,
  !    and return its unit number, setexist=.true.
  !    If the namelist NML isn't found, set exist=.false.

! MAB> -- needed for trinity
  public :: init_nt_name
  ! subroutine ...

  public :: nt_input_unit
  ! function nt_input_unit (nml)
  ! character(*), intent (in) :: nml
  ! integer :: nt_input_unit
  !    Rewind the Trinity input file to start of namelist NML,
  !    and return its unit number

  public :: nt_input_unit_exist
  ! function nt_input_unit_exist (nml,exist)
  ! character(*), intent (in) :: nml
  ! integer :: input_unit
  !    Rewind the Trinity input file to start of namelist NML,
  !    and return its unit number, setexist=.true.
  !    If the namelist NML isn't found, set exist=.false.

  public :: init_error_unit
  public :: init_input_unit, init_nt_input_unit
  public :: nt_num_input_lines

! <MAB

  public :: error_unit
  ! function error_unit ()
  ! integer :: error_unit
  !    Return the error unit number

  public :: get_input_unit

  public :: open_output_file
  ! subroutine open_output_file (unit, ext)
  ! integer, intent (out) :: unit
  ! character (*), intent (in) :: ext
  !    Open a file with name made from the run_name with the EXT appended
  !    and return its unit number in UNIT

  public :: mp_open_output_file
  ! subroutine mp_open_output_file (unit, ext, iproc)
  ! integer, intent (out) :: unit
  ! character (*), intent (in) :: ext
  ! integer, intent (in) :: iproc
  !    Open file on current processor with name made from list_name (was trin_name)
  !    (without .trin) and EXT appended, and return its unit number in UNIT

  public :: close_output_file
  ! subroutine close_output_file (unit)
  ! integer, intent (in) :: unit
  !    Close the file associated with UNIT from open_output_file

  public :: flush_output_file
  ! subroutine flush_output_file (unit)
  ! integer, intent (in) :: unit
  !    Close/open-append the file associated with UNIT from open_output_file

  public :: get_unused_unit
  ! subroutine get_unused_unit (unit)
  ! integer, intent (out) :: unit
  !    Return a unit number not associated with any file

  public :: get_indexed_namelist_unit
  ! subroutine get_indexed_namelist_unit (unit, nml, index)
  ! integer, intent (out) :: unit
  ! character (*), intent (in) :: nml
  ! integer, intent (in) :: index
  !    Copy namelist, NML // '_' // INDEX, from the input file to
  !    namelist, NML, in a temporary file, UNIT

  public :: num_input_lines

  public :: stdout_unit

  ! trin_name, nttrim, and nt_input_unit_no needed for Trinity -- MAB
  character (500), pointer :: run_name, nt_name
  character (500), target :: arun_name, job_name, nttrim
  character (500) :: list_name!, trin_name
  logical :: trin
  integer, save :: input_unit_no, error_unit_no=6, nt_input_unit_no
  integer, parameter :: stdout_unit=6
! TT>
!  integer, save, public :: num_input_lines
! nt_num_input_lines needed for Trinity -- MAB
  integer, save :: num_input_lines, nt_num_input_lines
! <TT

contains

  subroutine init_file_utils (list, input, error, name)
    implicit none
    logical, intent (out) :: list
    logical, intent (in), optional :: input, error
    character(*), intent (in), optional :: name
    logical :: inp, err

    if (present (input)) then
       inp = input
    else
       inp = .true.
    end if
    if (present (error)) then
       err = error
    else
       err = .true.
    end if
    if (present (name)) then
!# if FCOMPILER == _XL_
!       arun_name = name
!# else
       arun_name = trim(name)
!# endif
    else
       arun_name = "unknown"
    end if

! TT> changed for slice_g
!    call run_type (list)
    if (inp) then
       ! get runname from command line and
       ! set list=T if input ends in ".list"
       ! set trin=T if input ends in ".trin"
       call run_type (list)
    else
       list = .false.
       trin = .false.
    end if
! <TT

!    if (list) then
    if (list .or. trin) then
       list_name = arun_name
!    else if (trin) then
!       trin_name = arun_name
    else
       call init_run_name
       call init_error_unit (err)
       call init_input_unit (inp)
    end if

  end subroutine init_file_utils

  subroutine run_type (list)
! TT> This line made tungsten compiler (NCSA) to hang up
!    use command_line
# if FCOMPILER == _GFORTRAN_
    use command_line, only: cl_getarg
# else
    use command_line, only: iargc, cl_getarg
# endif
! <TT
    implicit none
    logical, intent (out) :: list
    integer :: l, ierr

    ! get argument from command line and put in arun_name
    if (iargc() /= 0) then
       call cl_getarg (1, arun_name, 500, ierr)
       if (ierr /= 0) then
          print *, "Error getting run name."
       end if
    end if

    ! count number of characters in arun_name (excluding trailing space)
    l = len_trim (arun_name)
!!$# if FCOMPILER == _XL_
!    if arun_name ends in .trin, set trin = T
    trin = (l > 5 .and. arun_name(l-4:l) == ".trin")
!    if arun_name ends in .list, set list = T
    list = (l > 5 .and. arun_name(l-4:l) == ".list")
!!$# else
!!$    if (l>5) then
!!$       list = arun_name(l-4:l) == ".list"
!!$    else
!!$       list = .false.
!!$    end if
!!$# endif

  end subroutine run_type

  subroutine init_run_name

    implicit none
    integer :: l

    l = len_trim (arun_name)
    if (l > 3 .and. arun_name(l-2:l) == ".in") then
       arun_name = arun_name(1:l-3)
    end if
    run_name => arun_name

  end subroutine init_run_name

! TT>
!  subroutine init_job_name (njobs, group0, job_list)
!    use command_line
!    use mp
!    use mp, only: scope, subprocs, job
  subroutine init_job_name (jobname)
! <TT
    implicit none
! TT>
!    integer, intent (in) :: njobs
!    integer, intent (in), dimension(0:) :: group0
!    character (len=500), dimension(0:) :: job_list
    character (len=500), intent (in) :: jobname
! <TT
    logical :: err = .true., inp = .true.

! TT>
!    call scope (subprocs)
!    job_name = trim(job_list(job))
    job_name = trim(jobname)
! <TT
    run_name => job_name

! MAB>
! can't call these here because we're on all processors
! and only want to init on proc0.  instead called
! from within job_manage
!    call init_error_unit (err)
!    call init_input_unit (inp)
! <MAB

  end subroutine init_job_name

  subroutine get_unused_unit (unit)
    implicit none
    integer, intent (out) :: unit
! TT>
!    character(20) :: read, write, readwrite
    logical :: od
! <TT
    unit = 50
    do
! TT>
!       inquire (unit=unit, read=read, write=write, readwrite=readwrite)
!       if (read == "UNKNOWN" .and. write == "UNKNOWN" &
!            .and. readwrite == "UNKNOWN") &
!       then
!          return
!       end if
       ! TT: Can we do the same thing like this?
       ! TT: The above fails in gfortran/g95.
       inquire (unit=unit, opened=od)
       if (.not.od) return
! <TT
       unit = unit + 1
    end do
  end subroutine get_unused_unit

  subroutine open_output_file (unit, ext)
    implicit none
    integer, intent (out) :: unit
    character (*), intent (in) :: ext
    character (500) :: hack
    call get_unused_unit (unit)
    hack=trim(run_name)//ext
    open (unit=unit, file=trim(hack), status="replace", action="write")
  end subroutine open_output_file

  subroutine mp_open_output_file (unit, ext, iproc)
    implicit none
    integer, intent (out) :: unit
    character (*), intent (in) :: ext
    integer, intent (in) :: iproc
    character (500) :: hack, proc
    call get_unused_unit (unit)
    write (proc,*) iproc
    proc = adjustl(proc)
!    hack=trin_name(:len_trim(trin_name)-5)//trim(proc)//ext
    hack=list_name(:len_trim(list_name)-5)//trim(proc)//ext
    open (unit=unit, file=trim(hack), status="replace", action="write")
  end subroutine mp_open_output_file

  subroutine close_output_file (unit)
    implicit none
    integer, intent (in) :: unit
    close (unit=unit)
  end subroutine close_output_file

  subroutine flush_output_file (unit, ext)
    implicit none
    integer, intent (in) :: unit
    character (*), intent (in), optional :: ext
    character (len=500) :: fname
    inquire(unit,name=fname)
# if FCOMPILER == _XL_
    call flush_(unit)
# elif FCOMPILER == _NAG_
    close(unit=unit)
    open (unit=unit, file=trim(fname), status="old", action="write", position="append")
# else
    call flush (unit)
# endif
  end subroutine flush_output_file

  subroutine init_error_unit (open_it)
    implicit none
    logical, intent (in) :: open_it
! TT> changed for slice_g
!    error_unit_no = 6
    error_unit_no = 0
! <TT
    if (run_name /= "unknown" .and. open_it) then
       call open_output_file (error_unit_no, ".error")
       ! TT: error_unit_no is overwritten for .error file
    end if
  end subroutine init_error_unit

  subroutine strip_comments (line)
    implicit none
    character(*), intent (in out) :: line
    logical :: in_single_quotes, in_double_quotes
    integer :: i, length

    length = len_trim(line)
    i = 1
    in_single_quotes = .false.
    in_double_quotes = .false.
    loop: do
       if (in_single_quotes) then
          if (line(i:i) == "'") in_single_quotes = .false.
       else if (in_double_quotes) then
          if (line(i:i) == '"') in_double_quotes = .false.
       else
          select case (line(i:i))
          case ("'")
             in_single_quotes = .true.
          case ('"')
             in_double_quotes = .true.
          case ("!")
             i = i - 1
             exit loop
          end select
       end if
       if (i >= length) exit loop
       i = i + 1
    end do loop
    line = line(1:i)
  end subroutine strip_comments

  subroutine init_input_unit (open_it)
    implicit none
    logical, intent (in) :: open_it
    integer :: in_unit, out_unit, iostat
    character(500) :: line

    ! for includes
    integer, parameter :: stack_size = 10
    integer, dimension (stack_size) :: stack
    integer :: stack_ptr

    if (.not. open_it) then
       input_unit_no = -1
       return
    end if

    call get_unused_unit (in_unit)
    open (unit=in_unit, file=trim(run_name)//".in", status="old", &
         action="read", iostat=iostat)
    if (iostat /= 0) then
       print "(a)", "Couldn't open input file: "//trim(run_name)//".in"
    end if

    call get_unused_unit (out_unit)
!    open (unit=out_unit, status="scratch", action="readwrite")
    open (unit=out_unit, file="."//trim(run_name)//".in")

    iostat = 0
    stack_ptr = 0
    num_input_lines = 0
    do
       read (unit=in_unit, fmt="(a)", iostat=iostat) line
       if (iostat /= 0) then
          if (stack_ptr <= 0) exit
          close (unit=in_unit)
          iostat = 0
          in_unit = stack(stack_ptr)
          stack_ptr = stack_ptr - 1
          cycle
       end if
       if (line(1:9) == "!include ") then
          if (stack_ptr >= stack_size) then
             print "(a)", "!include ignored: nesting too deep: "//trim(line)
             cycle
          end if
          stack_ptr = stack_ptr + 1
          stack(stack_ptr) = in_unit
          call get_unused_unit (in_unit)
          open (unit=in_unit, file=trim(line(10:)), status="old", &
                action="read", iostat=iostat)
          if (iostat /= 0) then
             print "(a)", "!include ignored: file unreadable: "//trim(line)
             in_unit = stack(stack_ptr)
             stack_ptr = stack_ptr - 1
             cycle
          end if
          cycle
       end if
       call strip_comments (line)
       write (unit=out_unit, fmt="(a)") trim(line)
       num_input_lines = num_input_lines + 1
    end do
    close (unit=in_unit)

    input_unit_no = out_unit
  end subroutine init_input_unit

  subroutine finish_file_utils
    implicit none
    if (input_unit_no > 0) then
       close (unit=input_unit_no)
       input_unit_no = -1
    end if
    if (error_unit_no > 0 .and. error_unit_no /= 6) then
       close (unit=error_unit_no)
       error_unit_no = -1
    end if
  end subroutine finish_file_utils

  function input_unit (nml)
    implicit none
    character(*), intent (in) :: nml
    integer :: input_unit, iostat
    character(500) :: line
    intrinsic adjustl, trim
    input_unit = input_unit_no
    if (input_unit_no > 0) then
       rewind (unit=input_unit_no)
       do
          read (unit=input_unit_no, fmt="(a)", iostat=iostat) line
          if (iostat /= 0) then
             rewind (unit=input_unit_no)
             exit
          end if
          if (trim(adjustl(line)) == "&"//nml) then
             backspace (unit=input_unit_no)
             return
          end if
       end do
    end if
    write (unit=error_unit_no, fmt="('Couldn''t find namelist: ',a)") nml
    write (unit=*, fmt="('Couldn''t find namelist: ',a)") nml
  end function input_unit

  function input_unit_exist (nml,exist)
    implicit none
    character(*), intent (in) :: nml
    logical, intent(out) :: exist
    integer :: input_unit_exist, iostat
    character(500) :: line
    intrinsic adjustl, trim
    input_unit_exist = input_unit_no
    exist = .true.
    if (input_unit_no > 0) then
       rewind (unit=input_unit_no)
       do
          read (unit=input_unit_no, fmt="(a)", iostat=iostat) line
          if (iostat /= 0) then
             rewind (unit=input_unit_no)
             exit
          end if
          if (trim(adjustl(line)) == "&"//nml) then
             backspace (unit=input_unit_no)
             return
          end if
       end do
    end if
    exist = .false.
  end function input_unit_exist

  function error_unit ()
    implicit none
    integer :: error_unit
    error_unit = error_unit_no
  end function error_unit

  subroutine get_input_unit (unit)
    implicit none
    integer, intent (out) :: unit

    unit = input_unit_no

  end subroutine get_input_unit

  subroutine get_indexed_namelist_unit (unit, nml, index)
    implicit none
    integer, intent (out) :: unit
    character (*), intent (in) :: nml
    integer, intent (in) :: index
    character(500) :: line
    integer :: iunit, iostat, in_file
    logical :: exist

    call get_unused_unit (unit)
!    open (unit=unit, status="scratch", action="readwrite")
    open (unit=unit, file="."//trim(run_name)//".scratch")

    write (line, *) index
    line = nml//"_"//trim(adjustl(line))
    in_file = input_unit_exist(trim(line), exist)
    if (exist) then
       iunit = input_unit(trim(line))
    else
       return
    end if

    read (unit=iunit, fmt="(a)") line
    write (unit=unit, fmt="('&',a)") nml

    do
       read (unit=iunit, fmt="(a)", iostat=iostat) line
       if (iostat /= 0 .or. trim(adjustl(line)) == "/") exit
       write (unit=unit, fmt="(a)") trim(line)
    end do
    write (unit=unit, fmt="('/')")
    rewind (unit=unit)
  end subroutine get_indexed_namelist_unit

! MAB> -- needed for Trinity
  subroutine init_nt_name (ntname)

    implicit none

    character (len=500), intent (in) :: ntname

    nttrim = trim(ntname)
    nt_name => nttrim

  end subroutine init_nt_name

  subroutine init_nt_input_unit (open_it)
    implicit none
    logical, intent (in) :: open_it
    integer :: in_unit, out_unit, iostat
    character(500) :: line

    ! for includes
    integer, parameter :: stack_size = 10
    integer, dimension (stack_size) :: stack
    integer :: stack_ptr

    if (.not. open_it) then
       nt_input_unit_no = -1
       return
    end if

    call get_unused_unit (in_unit)
!    open (unit=in_unit, file=trim(nt_name)//".trin", status="old", &
    open (unit=in_unit, file=trim(nt_name), status="old", &
         action="read", iostat=iostat)
    if (iostat /= 0) then
!       print "(a)", "Couldn't open input file: "//trim(nt_name)//".trin"
       print "(a)", "Couldn't open input file: "//trim(nt_name)
    end if

    call get_unused_unit (out_unit)
!    open (unit=out_unit, status="scratch", action="readwrite")
!    open (unit=out_unit, file="."//trim(trin_name)//".in")
    open (unit=out_unit, file="."//trim(nt_name))

    iostat = 0
    stack_ptr = 0
    nt_num_input_lines = 0
    do
       read (unit=in_unit, fmt="(a)", iostat=iostat) line
       if (iostat /= 0) then
          if (stack_ptr <= 0) exit
          close (unit=in_unit)
          iostat = 0
          in_unit = stack(stack_ptr)
          stack_ptr = stack_ptr - 1
          cycle
       end if
       if (line(1:9) == "!include ") then
          if (stack_ptr >= stack_size) then
             print "(a)", "!include ignored: nesting too deep: "//trim(line)
             cycle
          end if
          stack_ptr = stack_ptr + 1
          stack(stack_ptr) = in_unit
          call get_unused_unit (in_unit)
          open (unit=in_unit, file=trim(line(10:)), status="old", &
                action="read", iostat=iostat)
          if (iostat /= 0) then
             print "(a)", "!include ignored: file unreadable: "//trim(line)
             in_unit = stack(stack_ptr)
             stack_ptr = stack_ptr - 1
             cycle
          end if
          cycle
       end if
       call strip_comments (line)
       write (unit=out_unit, fmt="(a)") trim(line)
       nt_num_input_lines = nt_num_input_lines + 1
    end do
    close (unit=in_unit)

    nt_input_unit_no = out_unit
  end subroutine init_nt_input_unit

  function nt_input_unit (nml)
    implicit none
    character(*), intent (in) :: nml
    integer :: nt_input_unit, iostat
    character(500) :: line
    intrinsic adjustl, trim
    nt_input_unit = nt_input_unit_no
    if (nt_input_unit_no > 0) then
       rewind (unit=nt_input_unit_no)
       do
          read (unit=nt_input_unit_no, fmt="(a)", iostat=iostat) line
          if (iostat /= 0) then
             rewind (unit=nt_input_unit_no)
             exit
          end if
          if (trim(adjustl(line)) == "&"//nml) then
             backspace (unit=nt_input_unit_no)
             return
          end if
       end do
    end if
    write (unit=*, fmt="('Couldn''t find namelist: ',a)") nml
  end function nt_input_unit

  function nt_input_unit_exist (nml,exist)
    implicit none
    character(*), intent (in) :: nml
    logical, intent(out) :: exist
    integer :: nt_input_unit_exist, iostat
    character(500) :: line
    intrinsic adjustl, trim
    nt_input_unit_exist = nt_input_unit_no
    exist = .true.
    if (nt_input_unit_no > 0) then
       rewind (unit=nt_input_unit_no)
       do
          read (unit=nt_input_unit_no, fmt="(a)", iostat=iostat) line
          if (iostat /= 0) then
             rewind (unit=nt_input_unit_no)
             exit
          end if
          if (trim(adjustl(line)) == "&"//nml) then
             backspace (unit=nt_input_unit_no)
             return
          end if
       end do
    end if
    exist = .false.
  end function nt_input_unit_exist

! <MAB

end module file_utils
